using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;

// CustomUnityAsset Color Changer with tweaks by Blazedust. Original can be found here:
// https://www.reddit.com/r/VAMscenes/comments/a2mwi4/script_customunityasset_color_change_3axiz_scaling/

namespace DillDoe
{
    public class ColorScale : MVRScript
    {

        public string pluginAuthor = "DillDoe with modifications by Blazedust";
        public string pluginName = "ColorScaleC";
        public string pluginVersion = "1.4";
        public string pluginDate = "2020-06-07";
        public string pluginDescription = @"
        This plugin allows you to change the material color and size of the custom assets.  

        You are free to edit/change everything below this, please do not delete the PluginInfo. Append it if you modify the script 
		
		Edits by Blazedust to make all Renderer components in the GameObject to be affected without having to rely on an internal list in the plugin.";
		
        protected JSONStorableColor dcolor;
        protected JSONStorableColor scolor;
        protected JSONStorableFloat gfloat;
		protected JSONStorableFloat mfloat;
        protected JSONStorableFloat afloat;
        protected JSONStorableFloat jfloatX;
        protected JSONStorableFloat jfloatY;
        protected JSONStorableFloat jfloatZ;

		protected JSONStorableFloat dcfloatH;
		protected JSONStorableFloat dcfloatS;
		protected JSONStorableFloat dcfloatV;

		protected JSONStorableFloat scfloatH;
		protected JSONStorableFloat scfloatS;
		protected JSONStorableFloat scfloatV;

		protected JSONStorableFloat theRenderer;
		protected JSONStorableFloat theChild;
		protected JSONStorableFloat jfloatAll;
		protected JSONStorableBool scaleComponents;
		// protected JSONStorableString jstring;
		protected UIDynamicSlider dslider;
        // protected UIDynamicTextField dtext;
		protected int loadWaiter = 0;
		protected int mIndex = -1;
		protected int mRendererIndex = -1;
		protected JSONStorableBool[] affectToggles;
		protected UIDynamicToggle[] affectTogglesGui;

		protected JSONStorableStringChooser shaderTypeSelect;

		private Dictionary<GameObject, Vector3> originalMainScales;
		
		// 200 is plenty for smaller and bigger objects!
		const int MAX_COMPONENT_TOGGLES = 200;
		
		protected List<GOComp> comps;

		private Dictionary<string, ShaderInfo> shaderInfo;

		public class GOComp {
			
			public GameObject go;
			public Vector3 origScale;
			public int componentIndex;

			public JSONStorableBool affectToggle;
			
			public GOComp(GameObject g, int index)
			{
				go = g;
				componentIndex = index;
				if (go.transform.parent != null)
				{
					origScale = go.transform.parent.transform.localScale;
				} 
				else
                {
					origScale = new Vector3(1f, 1f, 1f);
                }
				affectToggle = null;
			}
		}

		class ShaderInfo
        {
			public string shaderType = "";
			public string renderMode = "";
			public Dictionary<Renderer, Material> materials;
			public ShaderInfo(string st, string rm)
            {
				shaderType = st;
				renderMode = rm;
				materials = new Dictionary<Renderer, Material>();
            }
		}

		public bool ComponentIncluded(GOComp c)
		{
			if (c.affectToggle != null && !c.affectToggle.val) {
				return false;
			}
			return true;
		}

        public override void Init()
        {
            try
            {				
				comps = new List<GOComp>();
				originalMainScales = new Dictionary<GameObject, Vector3>();

				// ************************************  Left Side  *************************************************************

				// Diffuse Color
				HSVColor hsvc = HSVColorPicker.RGBToHSV(1f, 1f, 1f);
                dcolor = new JSONStorableColor("DiffuseColor", hsvc, DiffuseColorCallback);
                // RegisterColor(dcolor);
                CreateColorPicker(dcolor);
				
                // Specular Color
                scolor = new JSONStorableColor("SpecularColor", hsvc, SpecColorCallback);
                // RegisterColor(scolor);
                CreateColorPicker(scolor);

				dcfloatH = new JSONStorableFloat("DiffuseColorH", 1f, DiffuseColorHCallback, 0f, 1f);
				RegisterFloat(dcfloatH);
				dcfloatS = new JSONStorableFloat("DiffuseColorS", 1f, DiffuseColorSCallback, 0f, 1f);
				RegisterFloat(dcfloatS);
				dcfloatV = new JSONStorableFloat("DiffuseColorV", 1f, DiffuseColorVCallback, 0f, 1f);
				RegisterFloat(dcfloatV);

				scfloatH = new JSONStorableFloat("SpecularColorH", 0f, SpecularColorHCallback, 0f, 1f);
				RegisterFloat(scfloatH);
				scfloatS = new JSONStorableFloat("SpecularColorS", 0f, SpecularColorSCallback, 0f, 1f);
				RegisterFloat(scfloatS);
				scfloatV = new JSONStorableFloat("SpecularColorV", 0f, SpecularColorVCallback, 0f, 1f);
				RegisterFloat(scfloatV);

				shaderInfo = new Dictionary<string, ShaderInfo>();
				shaderInfo.Add("Default", new ShaderInfo("Default", ""));
				shaderInfo.Add("Standard (Unity, Opaque)", new ShaderInfo("Standard", "Opaque"));
				shaderInfo.Add("Standard (Unity, Fade)", new ShaderInfo("Standard", "Fade"));
				shaderInfo.Add("Specular (Unity)", new ShaderInfo("Specular", "Opaque"));

				List<string> options = new List<string>();
				options.Add("Default");
				options.Add("Standard (Unity, Opaque)");
				options.Add("Standard (Unity, Fade)");
				options.Add("Specular (Unity)");
				shaderTypeSelect = new JSONStorableStringChooser("ShaderT", options, "Default", "Shader", (string shaderType) => {
					ShaderTypeCallback(shaderType);
					// FindChild();
					RendererCallback(theRenderer);
					ChildCallback(theChild);
					

					DiffuseColorCallback(dcolor);
					SpecColorCallback(scolor);
					GlossyCallback(gfloat);
					MetallicCallback(mfloat);
					AlphaCallback(afloat);
				});
				// UIDynamicPopup popup = CreatePopup(shaderTypeSelect, false); // WIP: only handles ONE material ATM, needs more work.

				// Spec
				gfloat = new JSONStorableFloat("Glossiness", 0.2f, GlossyCallback, 0f, 1f, true);
                RegisterFloat(gfloat);
                dslider = CreateSlider(gfloat);
				
				//  Metallic
				mfloat = new JSONStorableFloat("Metallic", 0f, MetallicCallback, 0f, 1f, true);
                RegisterFloat(mfloat);
                dslider = CreateSlider(mfloat);

                //  Alpha
                afloat = new JSONStorableFloat("Alpha", 1f, AlphaCallback, 0f, 1f, true);
                RegisterFloat(afloat);
                dslider = CreateSlider(afloat);


				// ************************************  Right Side  *************************************************************

				// Scale X
				JSONStorableFloat jfloatAll = new JSONStorableFloat("ScaleAll", 1f, ScaleAllCallback, -1f, 2f, false, true);
				CreateSlider(jfloatAll, true);

				// Scale X
				jfloatX = new JSONStorableFloat("ScaleX", 1f, ScaleX, -1f, 2f, false, true);
                //jfloatX.storeType = JSONStorableParam.StoreType.Full;
                RegisterFloat(jfloatX);
                CreateSlider(jfloatX, true);

                // Scale Y
                jfloatY = new JSONStorableFloat("ScaleY", 1f, ScaleY, -1f, 2f, false, true);
                RegisterFloat(jfloatY);
                CreateSlider(jfloatY, true);

                // Scale Z
                jfloatZ = new JSONStorableFloat("ScaleZ", 1f, ScaleZ, -1f, 2f, false, true);
                RegisterFloat(jfloatZ);
                CreateSlider(jfloatZ, true);

				scaleComponents = new JSONStorableBool("ScaleComponents", false, ScaleComponentsCallback);
				RegisterBool(scaleComponents);
				UIDynamicToggle toggle = CreateToggle(scaleComponents, true);
				toggle.label = "Scale Components (On)\nWhole Object (Off)";
				toggle.height = 80;

				/*UIDynamic spacer = CreateSpacer(true);
                spacer.height = 240f;
                */

				// jstring = new JSONStorableString("Childs", "");
				// dtext = CreateTextField(jstring,true);
				// if (go != null)
				// {
				//     FindChild();
				// }

				theRenderer = new JSONStorableFloat("Renderer Index (-1 = All)", -1f, RendererCallback, -1f, 10);
				RegisterFloat(theRenderer);
				dslider = CreateSlider(theRenderer, true);
				dslider.slider.wholeNumbers = true;

				theChild = new JSONStorableFloat("Render Material Index (-1 = All)", -1f, ChildCallback, -1f, 10);
                RegisterFloat(theChild);
                dslider = CreateSlider(theChild, true);
                dslider.slider.wholeNumbers = true;
				
				// Pre-create toggle-boxes over which ones should affect and not.
				affectToggles = new JSONStorableBool[MAX_COMPONENT_TOGGLES];
				affectTogglesGui = new UIDynamicToggle[affectToggles.Length];
				for (int i = 0; i < affectToggles.Length; i ++) 
				{
					affectToggles[i] = new JSONStorableBool("component_" + i.ToString(), false);
					RegisterBool(affectToggles[i]);
				}

			}
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }

		void OnDestroy()
        {
			RemoveColorPicker(dcolor);
			RemoveColorPicker(scolor);
			RemoveSlider(jfloatAll);
			RemoveSlider(gfloat);
			RemoveSlider(mfloat);
			RemoveSlider(afloat);
			RemoveSlider(jfloatX);
			RemoveSlider(jfloatY);
			RemoveSlider(jfloatZ);
			RemoveSlider(theChild);
			RemoveToggle(scaleComponents);

			for (int i = 0; i < affectToggles.Length; i ++)
            {
				if (affectToggles[i] != null)
				{
					RemoveToggle(affectToggles[i]);
				}
			}
		}

			private void InitComponentsGUI() 
		{
			
			UIDynamicButton refreshAll = CreateButton("Refresh selected components", true);
            refreshAll.button.onClick.AddListener(() =>
            {
				try {
					LoadVal();
				} catch (Exception e) {
					SuperController.LogError(e.ToString());
				}
            });
			refreshAll.buttonColor = new Color(0.7f, 0.7f, 0.9f);
			
			// Create a "Select all" and "Deselect all" button.
			UIDynamicButton selectAll = CreateButton("Select All Components to affect", true);
            selectAll.button.onClick.AddListener(() =>
            {
				try {
					SetAllComponents(true);
					LoadVal();
				} catch (Exception e) {
					SuperController.LogError(e.ToString());
				}
            });
			selectAll.buttonColor = new Color(0.7f, 0.9f, 0.7f);
			
			UIDynamicButton deselectAll = CreateButton("Deselect All Components", true);
            deselectAll.button.onClick.AddListener(() =>
            {
				try {
					SetAllComponents(false);
				} catch (Exception e) {
					SuperController.LogError(e.ToString());
				}
            });
			deselectAll.buttonColor = new Color(0.9f, 0.7f, 0.7f);
			
			int max = Math.Min(comps.Count, affectToggles.Length);
			for (int i = 0; i < max; i ++) 
			{
				affectTogglesGui[i] = CreateToggle(affectToggles[i], true);
				affectTogglesGui[i].label = i.ToString() + ":" + comps[i].go.name;
				affectTogglesGui[i].backgroundColor = Color.yellow;
				comps[i].affectToggle = affectToggles[i]; 
			}
		}

		private void SetAllComponents(bool val)
		{
			for (int i = 0; i < affectToggles.Length; i ++) 
			{
				if (affectToggles[i] != null) 
				{
					affectToggles[i].val = val;
				} else {
					// reached end of toggles
					break;
				}
			}
		}

        void Update()
        {
			if (loadWaiter == 200 || SuperController.singleton.isLoading) 
			{
				return;
			}
            try
            {
				//  There seems to be a slight delay that prevents finding gameobject on load,
				//  so wait for 200 update cycles after main loading is over before trying.

				loadWaiter += 1;
				if (loadWaiter == 200) 
				{
					comps = new List<GOComp>();
					
					Atom atom = GetContainingAtom();
					
					int componentIndex = 0;
					foreach (Transform t in atom.onToggleObjects) 
					{
						if (t != null && t.gameObject != null) 
						{
							originalMainScales[t.gameObject] = t.gameObject.transform.localScale;

							Component[] components = t.gameObject.GetComponentsInChildren(typeof(Renderer)); // SkinnedMeshRenderer, MeshRenderer;
							//Component[] components = t.gameObject.GetComponentsInChildren(typeof(MeshRenderer)); // SkinnedMeshRenderer, MeshRenderer;
							//Component[] components = t.gameObject.GetComponentsInChildren(typeof(SkinnedMeshRenderer)); // SkinnedMeshRenderer, MeshRenderer;
							if (components != null && components.Length > 0) 
							{
								for (int k = 0; k < components.Length; k ++) 
								{
									// SuperController.LogMessage("component " + componentIndex.ToString() + ":" + components[k].name + ", " + components[k].GetType().ToString());	
									comps.Add(new GOComp(components[k].gameObject, componentIndex));
									componentIndex ++;
								}
							}

						}
					}
					
					if (comps.Count > 0)
					{
						// SuperController.LogMessage("ColorScale: " + comps.Count.ToString() + " components to load");
						InitComponentsGUI();
						LoadVal();
					} 
					else 
					{
						SuperController.LogMessage("ColorScale: Nothing to load");
					}
				}
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }

        //  Load all the saved values from plugin UI back into Asset
        void LoadVal()
        {
            try
            {
				// ShaderTypeCallback(shaderTypeSelect.val);

				ScaleX(jfloatX);
                ScaleY(jfloatY);
                ScaleZ(jfloatZ);

				RendererCallback(theRenderer);
                ChildCallback(theChild);

                DiffuseColorCallback(dcolor);
                SpecColorCallback(scolor);
                GlossyCallback(gfloat);
                MetallicCallback(mfloat);
                AlphaCallback(afloat);

            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }

		protected void ShaderTypeCallback(string shaderType)
        {
			ShaderInfo sInfo = shaderInfo[shaderType];

			ComponentRenderers().ForEach((r) => {

				try
				{
					if (!shaderInfo["Default"].materials.ContainsKey(r))
					{
						shaderInfo["Default"].materials.Add(r, r.material);
					}

					if (sInfo.shaderType == "Default")
					{
						Material dMat = null;
						if (shaderInfo["Default"].materials.TryGetValue(r, out dMat))
						{
							// SuperController.LogMessage("Reverting back to default");
							r.material = dMat;
						}
						else
						{
							// SuperController.LogMessage("Default already exist");
							// already having the default material
						}
					}
					else
					{
						Material mat = null;
						Material matB = r.material;

						if (!shaderInfo["Default"].materials.ContainsKey(r))
						{
							shaderInfo["Default"].materials.Add(r, r.material);
						}

						if (!sInfo.materials.TryGetValue(r, out mat))
						{
							// SuperController.LogMessage("Creating new: " + shaderType);
							mat = new Material(Shader.Find(sInfo.shaderType));
							sInfo.materials.Add(r, mat);

							mat.CopyPropertiesFromMaterial(matB);

							// mat.SetTexture("_MainTex", matB.GetTexture("_BumpMap")); // Redish texture!?

							// mat.SetTexture("_BumpMap", matB.GetTexture("_MainTex"));
							// mat.SetTexture("_DetailNormalMap", matB.GetTexture("_MainTex"));
							// mat.SetTexture("_DetailNormalMap", matB.GetTexture("_BumpMap"));

							// mat.SetTexture("_MainTex", matB.GetTexture("_MainTex"));
							// mat.SetFloat("_AlphaAdjust", matB.GetFloat("_AlphaAdjust"));
							// mat.SetColor("_Color", matB.GetColor("_Color"));
							//mat.mainTexture = matB.mainTexture;
							//mat.mainTextureOffset = matB.mainTextureOffset;
							//mat.mainTextureScale = matB.mainTextureScale;
							//mat.color = matB.color;

							//mat.SetFloat("_Shininess", matB.GetFloat("_Shininess"));
							//mat.SetFloat("_Cutoff", matB.GetFloat("_Cutoff"));
							//mat.SetFloat("_Glossiness", matB.GetFloat("_Glossiness"));
							//mat.SetFloat("_GlossMapScale", matB.GetFloat("_GlossMapScale"));
							//mat.SetFloat("_SmoothnessTextureChannel", matB.GetFloat("_SmoothnessTextureChannel"));
							//mat.SetFloat("_Metallic", matB.GetFloat("_Metallic"));
							//mat.SetTexture("_MetallicGlossMap", matB.GetTexture("_MetallicGlossMap"));
							//mat.SetTextureOffset("_MetallicGlossMap", matB.GetTextureOffset("_MetallicGlossMap"));
							//mat.SetTextureScale("_MetallicGlossMap", matB.GetTextureScale("_MetallicGlossMap"));
							//mat.SetFloat("_SpecularHighlights", matB.GetFloat("_SpecularHighlights"));
							//mat.SetFloat("_GlossyReflections", matB.GetFloat("_GlossyReflections"));

							//mat.SetTexture("_BumpMap", matB.GetTexture("_BumpMap")); // Just doesn't seem to work!?
							//mat.SetTextureOffset("_BumpMap", matB.GetTextureOffset("_BumpMap"));
							//mat.SetTextureScale("_BumpMap", matB.GetTextureScale("_BumpMap"));
							//mat.SetFloat("_BumpScale", matB.GetFloat("_BumpScale"));
							//mat.SetFloat("_Parallax", matB.GetFloat("_Parallax"));
							//mat.SetTexture("_ParallaxMap", matB.GetTexture("_ParallaxMap"));
							//mat.SetTextureOffset("_ParallaxMap", matB.GetTextureOffset("_ParallaxMap"));
							//mat.SetTextureScale("_ParallaxMap", matB.GetTextureScale("_Bum_ParallaxMappMap"));
							//mat.SetFloat("_OcclusionStrength", matB.GetFloat("_OcclusionStrength"));
							//mat.SetTexture("_OcclusionMap", matB.GetTexture("_OcclusionMap"));
							//mat.SetTextureOffset("_OcclusionMap", matB.GetTextureOffset("_OcclusionMap"));
							//mat.SetTextureScale("_OcclusionMap", matB.GetTextureScale("_OcclusionMap"));
							//mat.SetColor("_EmissionColor", matB.GetColor("_EmissionColor"));
							//mat.SetTexture("_EmissionMap", matB.GetTexture("_EmissionMap"));
							//mat.SetTextureOffset("_EmissionMap", matB.GetTextureOffset("_EmissionMap"));
							//mat.SetTextureScale("_EmissionMap", matB.GetTextureScale("_EmissionMap"));
							//mat.SetTexture("_DetailMask", matB.GetTexture("_DetailMask"));
							//mat.SetTextureOffset("_DetailMask", matB.GetTextureOffset("_DetailMask"));
							//mat.SetTextureScale("_DetailMask", matB.GetTextureScale("_DetailMask"));
							//mat.SetTexture("_DetailAlbedoMap", matB.GetTexture("_DetailAlbedoMap"));
							//mat.SetTextureOffset("_DetailAlbedoMap", matB.GetTextureOffset("_DetailAlbedoMap"));
							//mat.SetTextureScale("_DetailAlbedoMap", matB.GetTextureScale("_DetailAlbedoMap"));

							//mat.SetTexture("_DetailNormalMap", matB.GetTexture("_DetailNormalMap"));
							//mat.SetTextureOffset("_DetailNormalMap", matB.GetTextureOffset("_DetailNormalMap"));
							//mat.SetTextureScale("_DetailNormalMap", matB.GetTextureScale("_DetailNormalMap"));
							//mat.SetFloat("_DetailNormalMapScale", matB.GetFloat("_DetailNormalMapScale"));
							//mat.SetFloat("_UVSec", matB.GetFloat("_UVSec"));
							//mat.SetFloat("_Mode", matB.GetFloat("_Mode"));
							//mat.SetFloat("_SrcBlend", matB.GetFloat("_SrcBlend"));
							//mat.SetFloat("_DstBlend", matB.GetFloat("_DstBlend"));
							//mat.SetFloat("_ZWrite", matB.GetFloat("_ZWrite"));

							foreach (string kw in r.material.shaderKeywords)
							{
								mat.EnableKeyword(kw);
							}
							SetMaterialKeywords(mat);
							// mat.EnableKeyword("_NORMALMAP");
							// mat.EnableKeyword("NORMALMAP");

							//Color c = new Color(1f, 0f, 1f);
							mat.EnableKeyword("_EMISSION");
							mat.SetColor("_EmissionColor", new Color(1f, 0f, 1f, 0.5f));
							mat.SetTexture("_EmissionMap", Texture2D.whiteTexture);

							SetupMaterialWithBlendMode(mat, sInfo.renderMode);
							
						}
						
						r.material = mat;
					}
				} 
				catch(Exception e)
                {
					SuperController.LogError(e.ToString());
                }
			});
		}

		private bool IsDefaultMaterial(Renderer r)
        {
			Material mat = null;
			return (shaderInfo["Default"].materials.TryGetValue(r, out mat) && r.material == mat);
        }

		//protected void RenderModeCallback(string renderMode)
		//{
		//	ComponentRenderers().ForEach((r) =>
		//	{
		//		try
		//		{
		//			if (!IsDefaultMaterial(r))
		//			{
		//				SetupMaterialWithBlendMode(r.material, renderMode);
		//			}
		//		}
		//		catch (Exception e)
		//		{
		//			SuperController.LogError(e.ToString());
		//		}
		//	});
		//}

		protected void ScaleComponentsCallback(JSONStorableBool jb)
        {
			ScaleX(jfloatX);
			ScaleY(jfloatY);
			ScaleZ(jfloatZ);
		}
		
		protected void ScaleAllCallback(JSONStorableFloat jAll)
        {
			jfloatX.val = jAll.val;
			jfloatY.val = jAll.val;
			jfloatZ.val = jAll.val;
		}

		protected void ScaleC(Func<Vector3, Vector3, Vector3> compScale)
		{
			Atom atom = GetContainingAtom();
			if (scaleComponents.val)
			{
				// Restore original
				foreach (Transform t in atom.onToggleObjects)
				{
					if (t != null && t.gameObject != null)
					{
						Vector3 originalScale = Vector3.zero;
						if (originalMainScales.TryGetValue(t.gameObject, out originalScale))
						{
							Transform trans = t.gameObject.transform;
							trans.localScale = originalScale;
						}
					}
				}

				// Apply new 
				comps.ForEach((x) =>
				{
					if (ComponentIncluded(x))
					{
						if (x.go.transform != null && x.go.transform.parent != null)
						{
							Transform trans = x.go.transform.parent.transform;
							if (trans != null)
							{
								trans.localScale = compScale(x.origScale, trans.localScale);
							}
						}
					}
				});
			}
			else
			{
				// Restore original
				comps.ForEach((x) =>
				{
					// if (ComponentIncluded(x))
					{
						if (x.go.transform != null && x.go.transform.parent != null)
						{
							Transform trans = x.go.transform.parent.transform;
							if (trans != null)
							{
								trans.localScale = x.origScale;
							}
						}
					}
				});

				// Apply new 
				foreach (Transform t in atom.onToggleObjects)
				{
					if (t != null && t.gameObject != null)
					{
						Vector3 originalScale = Vector3.zero;
						if (originalMainScales.TryGetValue(t.gameObject, out originalScale))
						{
							Transform trans = t.gameObject.transform;
							trans.localScale = compScale(originalScale, trans.localScale);
						}
					}
				}
			}
		}

		//	Scale		
		protected void ScaleX(JSONStorableFloat jX)
        {
			ScaleC((Vector3 origScale, Vector3 localScale) => { return new Vector3(jX.val * origScale.x, localScale.y, localScale.z); });
        }

        protected void ScaleY(JSONStorableFloat jY)
        {
			ScaleC((Vector3 origScale, Vector3 localScale) => { return new Vector3(localScale.x, jY.val * origScale.y, localScale.z); });
        }

        protected void ScaleZ(JSONStorableFloat jZ)
        {
			ScaleC((Vector3 origScale, Vector3 localScale) => { return new Vector3(localScale.x, localScale.y, jZ.val * origScale.z); });
		}

        //  find all the child object in prefab to change each material color
        // protected void FindChild()
        // {            
        //     if (go != null)
        //     {   // todo: create dynamic array to hold color info for each prefab to save with scene
        //         //Dictionary<string, HSVColor> prefabChild = new Dictionary<string, HSVColor>();
        //         float x = 0f;
        //         foreach (var ob in go.GetComponentsInChildren<Renderer>())
        //         {
        //             x++;
        //             jstring.val += x + ". " + ob.name + "\n";
        //         }             
        //     }
        // }
        // prefab selector
		protected void RendererCallback(JSONStorableFloat jC)
		{
			// SuperController.LogMessage("mRendererIndex pre " + mRendererIndex.ToString());
			mRendererIndex = Math.Max((int)jC.val, -1);
			// SuperController.LogMessage("mRendererIndex post " + mRendererIndex.ToString());
		}

		protected void ChildCallback(JSONStorableFloat jC)
        {
			// SuperController.LogMessage("mIndex pre " + mIndex.ToString());
            mIndex = Math.Max((int)jC.val, -1);
			// SuperController.LogMessage("mIndex post " + mIndex.ToString());
        }

		/// <summary>
        /// Gets renderer materials based on selected material index.
        /// </summary>
		private IEnumerable<Material> GetRendereMaterials(Renderer r)
        {
			if (r.materials != null && r.materials.Length > 0)
            {
				if (mIndex == -1)
                {
					// all
					foreach(Material mat in r.materials)
                    {
						yield return mat;
					}
				} 
				else if (mIndex < r.materials.Length)
                {
					yield return r.materials[mIndex];
                }
            }
		}

		/// <summary>
        /// Gets all renderes and materials in all included components to iterate through.
        /// </summary>
		private List<Material> ComponentRenderersMaterials()
        {
			List<Material> result = new List<Material>();
			comps.ForEach((x) =>
			{
				if (ComponentIncluded(x))
				{
					Renderer[] rends = x.go.GetComponentsInChildren<Renderer>();
					if (rends != null && rends.Length > 0)
                    {
						if (mRendererIndex == -1)
						{
							// all renderers
							foreach (Renderer rend in rends)
							{
								result.AddRange(GetRendereMaterials(rend));
							}
						}
						else if (mRendererIndex < rends.Length)
						{
							result.AddRange(GetRendereMaterials(rends[mRendererIndex]));
						}
					}
				}
			});
			return result;
		}

		private List<Renderer> ComponentRenderers()
		{
			List<Renderer> result = new List<Renderer>();
			comps.ForEach((x) =>
			{
				if (ComponentIncluded(x))
				{
					Renderer[] rends = x.go.GetComponentsInChildren<Renderer>();
					if (mRendererIndex == -1)
					{
						// all renderers
						result.AddRange(rends);
					}
					else if (mRendererIndex < rends.Length)
					{
						result.Add(rends[mRendererIndex]);
					}
				}
			});
			return result;
		}

		protected void DiffuseColorHCallback(JSONStorableFloat jf)
        {
			HSVColor hc = dcolor.val;
			hc.H = jf.val;
			dcolor.val = hc;
		}

		protected void DiffuseColorSCallback(JSONStorableFloat jf)
        {
			HSVColor hc = dcolor.val;
			hc.S = jf.val;
			dcolor.val = hc;
		}

		protected void DiffuseColorVCallback(JSONStorableFloat jf)
        {
			HSVColor hc = dcolor.val;
			hc.V = jf.val;
			dcolor.val = hc;
		}
		
		//  material color
		protected void DiffuseColorCallback(JSONStorableColor dc)
        {
			dcfloatH.val = dc.valNoCallback.H;
			dcfloatS.val = dc.valNoCallback.S;
			dcfloatV.val = dc.valNoCallback.V;
			Color c = Color.HSVToRGB(dc.valNoCallback.H, dc.valNoCallback.S, dc.valNoCallback.V);
			c.a = afloat.val;
			ComponentRenderersMaterials().ForEach((m) => {
				m.SetColor("_Color", c);
			});
		}

		protected void SpecularColorHCallback(JSONStorableFloat jf)
		{
			HSVColor hc = scolor.val;
			hc.H = jf.val;
			scolor.val = hc;
		}

		protected void SpecularColorSCallback(JSONStorableFloat jf)
		{
			HSVColor hc = scolor.val;
			hc.S = jf.val;
			scolor.val = hc;
		}

		protected void SpecularColorVCallback(JSONStorableFloat jf)
		{
			HSVColor hc = scolor.val;
			hc.V = jf.val;
			scolor.val = hc;
		}

		protected void SpecColorCallback(JSONStorableColor sc)
        {
			scfloatH.val = sc.valNoCallback.H;
			scfloatS.val = sc.valNoCallback.S;
			scfloatV.val = sc.valNoCallback.V;
			Color c = Color.HSVToRGB(sc.valNoCallback.H, sc.valNoCallback.S, sc.valNoCallback.V);
			ComponentRenderersMaterials().ForEach((m) => {
				m.SetColor("_SpecColor", c);
            });
        }

        protected void GlossyCallback(JSONStorableFloat jf)
        {
			ComponentRenderersMaterials().ForEach((m) => {
				m.SetFloat("_Glossiness", jf.val);
				m.SetFloat("_Shininess", jf.val);
            });
        }

        protected void MetallicCallback(JSONStorableFloat jm)
        {
			ComponentRenderersMaterials().ForEach((m) => {
				m.SetFloat("_Metallic", jm.val);
            });
        }

        protected void AlphaCallback(JSONStorableFloat ja)
        {
			ComponentRenderersMaterials().ForEach((m) => {
				Color colorA = m.color;
				colorA.a = ja.val;
				m.SetColor("_Color", colorA);
            });
        }

		// https://github.com/Unity-Technologies/UnityCsReference/blob/master/Editor/Mono/Inspector/StandardShaderGUI.cs
		void SetupMaterialWithBlendMode(Material material, string blendMode)
		{
			switch (blendMode)
			{
				case "Opaque":
					material.SetOverrideTag("RenderType", "");
					material.SetInt("_SrcBlend", (int)UnityEngine.Rendering.BlendMode.One);
					material.SetInt("_DstBlend", (int)UnityEngine.Rendering.BlendMode.Zero);
					material.SetInt("_ZWrite", 1);
					material.DisableKeyword("_ALPHATEST_ON");
					material.DisableKeyword("_ALPHABLEND_ON");
					material.DisableKeyword("_ALPHAPREMULTIPLY_ON");
					material.renderQueue = -1;
					break;
				case "Cutout":
					material.SetOverrideTag("RenderType", "TransparentCutout");
					material.SetInt("_SrcBlend", (int)UnityEngine.Rendering.BlendMode.One);
					material.SetInt("_DstBlend", (int)UnityEngine.Rendering.BlendMode.Zero);
					material.SetInt("_ZWrite", 1);
					material.EnableKeyword("_ALPHATEST_ON");
					material.DisableKeyword("_ALPHABLEND_ON");
					material.DisableKeyword("_ALPHAPREMULTIPLY_ON");
					material.renderQueue = (int)UnityEngine.Rendering.RenderQueue.AlphaTest;
					break;
				case "Fade":
					material.SetOverrideTag("RenderType", "Transparent");
					material.SetInt("_SrcBlend", (int)UnityEngine.Rendering.BlendMode.SrcAlpha);
					material.SetInt("_DstBlend", (int)UnityEngine.Rendering.BlendMode.OneMinusSrcAlpha);
					material.SetInt("_ZWrite", 0);
					material.DisableKeyword("_ALPHATEST_ON");
					material.EnableKeyword("_ALPHABLEND_ON");
					material.DisableKeyword("_ALPHAPREMULTIPLY_ON");
					material.renderQueue = (int)UnityEngine.Rendering.RenderQueue.Transparent;
					break;
				case "Transparent":
					material.SetOverrideTag("RenderType", "Transparent");
					material.SetInt("_SrcBlend", (int)UnityEngine.Rendering.BlendMode.One);
					material.SetInt("_DstBlend", (int)UnityEngine.Rendering.BlendMode.OneMinusSrcAlpha);
					material.SetInt("_ZWrite", 0);
					material.DisableKeyword("_ALPHATEST_ON");
					material.DisableKeyword("_ALPHABLEND_ON");
					material.EnableKeyword("_ALPHAPREMULTIPLY_ON");
					material.renderQueue = (int)UnityEngine.Rendering.RenderQueue.Transparent;
					break;
			}
		}

		void SetMaterialKeywords(Material material)
		{
			SetKeyword(material, "_NORMALMAP", material.GetTexture("_BumpMap") || material.GetTexture("_DetailNormalMap"));
			// SetKeyword(material, "_SPECGLOSSMAP", material.GetTexture("_SpecGlossMap"));
			SetKeyword(material, "_METALLICGLOSSMAP", material.GetTexture("_MetallicGlossMap"));
			SetKeyword(material, "_PARALLAXMAP", material.GetTexture("_ParallaxMap"));
			SetKeyword(material, "_DETAIL_MULX2", material.GetTexture("_DetailAlbedoMap") || material.GetTexture("_DetailNormalMap"));
			
			bool shouldEmissionBeEnabled = (material.globalIlluminationFlags & MaterialGlobalIlluminationFlags.EmissiveIsBlack) == 0;
			SetKeyword(material, "_EMISSION", shouldEmissionBeEnabled);
		}

		void SetKeyword(Material m, string keyword, bool state)
		{
			// SuperController.LogMessage(keyword + ": " + state.ToString());
			if (state)
				m.EnableKeyword(keyword);
			else
				m.DisableKeyword(keyword);
		}
	}
}